/*
*  Name:      application.hh
*  Author:    Rafael Jesus Alcantara Perez
*  Summary:   UESQLC application
*  Date:      $Date: 2000/02/20 12:15:32 $
*  Revision:  $Revision: 1.23 $
*
*  Copyright (C) 1996-2000  Rafael Jesus Alcantara Perez <rafa@dedalo-ing.com>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
*  MA 02111-1307, USA.
*/

#ifndef _UESQLC_APPLICATION__
#define _UESQLC_APPLICATION__

#include <cstdio>
#include <mpcl/base_application.hh>
#include <mpcl/config_processor.hh>
#include <mpcl/extended_map.hh>
#include <mpcl/pattern_based_code_generator.hh>
#include <mpcl/smart_pointer.hh>
#include <string>
#include "defs.hh"
#include "invariant_verifier.hh"


///  UESQLC compiler application.
class TApplication : public mpcl::TBaseApplication
{

  protected:

    ///  Smart pointer to pattern based code generator type definition.
    typedef 
      mpcl::TSmartPointer <mpcl::TPatternBasedCodeGenerator>  
      QTPatternBasedCodeGenerator;

    /**
    *  Map of target run-time name to smart pointer to pattern based
    *  code generator type.
    */
    typedef
      mpcl::xextended_map<std::string, QTPatternBasedCodeGenerator>
      xtarget_to_code_generator_map;

    ///  Map of target run-time name to variable maps type definition.
    typedef
      mpcl::xextended_map<std::string, mpcl::xvariable_map>
      xtarget_to_runtime_parameters_map_map;


  protected:

    ///  True if option "complementary-files" has been parsed in configuration stage.
    bool                                         gCfgComplementaryFiles;
    
    ///  True if option "help" has been parsed in configuration stage.
    bool                                         gCfgHelp;

    ///  True if option "line-directives" option has been parsed in configuration stage.
    bool                                         gCfgLineDirectives;

    ///  True if option "main-file" has been parsed in configuration stage.
    bool                                         gCfgMainFile;
    
    ///  True if option "output-cdml" has been parsed in configuration stage.
    bool                                         gCfgOutputCdml;

    ///  True if option "output-cdml-skel" option has been parsed in configuration stage.
    bool                                         gCfgOutputCdmlSkel;

    ///  True if option "parameters" has been parsed in configuration stage.
    bool                                         gCfgParameters;

    ///  True if option "version" option has been parsed in configuration stage.
    bool                                         gCfgVersion;

    ///  Compilation successfulness.
    bool                                         gCompilationOk;

    ///  True if the current target has been loaded before.
    bool                                         gReusingTarget;

    ///  Pointer to runtime variable map for code-generator.
    mpcl::xvariable_map*                         ptRuntimeParametersMap;

    ///  Variable map with arguments for code-generator.
    mpcl::xvariable_map                          tCodeParametersMap;

    ///  Configuration processor.
    mpcl::TConfigProcessor                       tConfigProcessor;

    ///  Invariant verifier for compiler.
    TInvariantVerifier                         tInvariantVerifier;

    /**
    *  Map with correspondences between reserved words character
    *  representation and, their token value.
    */
    mpcl::xextended_map<std::string, long int>   tReservedWordMap;

    /**
    *  Map of target run-time name to smart pointer to pattern based
    *  code generator.
    */
    xtarget_to_code_generator_map                tTargetToCodeGeneratorMap;

    ///  Map with target to runtime parameters map.
    xtarget_to_runtime_parameters_map_map        tTargetToRuntimeParametersMapMap;

    ///  Debug level from configuration stage.
    unsigned int                                 uiCfgDebug;

    ///  Path for CDML files.
    std::string                                  yCfgCdmlPath;

    ///  Name of the input file name.
    std::string                                  yCfgInputFileName;

    ///  Name of the output file name.
    std::string                                  yCfgOutputFileName;

    ///  Target platform.
    std::string                                  yCfgTarget;

    ///  Current target name.
    std::string                                  yCurrentTarget;

    ///  String with the last token read from scanner in symbolic form.
    std::string                                  yLastToken;

    ///  Output code.
    std::string                                  yOutputCode;

    ///  Postfix output code.
    std::string                                  yPostfixOutputCode;

    ///  Prefix output code.
    std::string                                  yPrefixOutputCode;

    ///  Line number for embedded introducer for UESQL in host language.
    size_t                                       zEmbeddedIntroducerLine;


  public:

    //
    //  C O N S T R U C T O R S
    //

    /**
    *  Builds a new instance.
    *  @param pkcAPPLICATION_NAME    Application name.
    *  @param pkcAPPLICATION_RELEASE Application release.
    */
    TApplication (const char* pkcAPPLICATION_NAME, const char* pkcAPPLICATION_RELEASE);

    ///  Destroys the instance.
    ~TApplication (void);

    /**
    *  Binds code parameter \a rktVARIABLE with value \a rkyVALUE.
    *  @param rkyVARIABLE Code parameter name.
    *  @param rkyVALUE    Code parameter value.
    */
    void bindCodeParameter (const std::string& rkyVARIABLE, const std::string& rkyVALUE);

    /**
    *  Collapses multiple string_literal parts into an only one.
    *  @param pkcSOURCE_TEXT Source text with multiple string_literal parts.
    *  @return Collapsed string.
    */
    std::string collapseStringLiteral (const char* pkcSOURCE_TEXT);
    
    /**
    *  Extracts target from \b EMBEDDED_SQL_BEGIN and sets the proper CDML target.
    *  @param pkcSOURCE_TEXT Source text with parsed string.
    */
    void handleTarget (const char* pkcSOURCE_TEXT);

    /**
    *  Searches  reserved-word map for  the token bound to \a  pkcSOURCE_TEXT,
    *  and  if  it  is   found,  then  the  token  is   returned,  else  token
    *  \b IDENTIFIER_BODY  is returned. But, if \a uiPERIOD_COUNT is not zero,
    *  then it calls splitToken() cause it is multi-identifier token.
    *  @param pkcSOURCE_TEXT Token name.
    *  @param uiPERIOD_COUNT Period count.
    *  @return Token code.
    */
    int identifyToken (const char* pkcSOURCE_TEXT, unsigned int uiPERIOD_COUNT = 0);

    /**
    *  Initializes  \a  tCodeParametersMap  with  one  value.  That  value  is
    *  "_target", that contains the current target value.
    */
    void initCodeParameters (void);

    /**
    *  Initialize from parameters.
    *  @param iPARAMETERS_COUNT   Main argc.
    *  @param ppkcPARAMETERS_LIST Main argv.
    */
    void initialize (int iPARAMETERS_COUNT, const char** ppkcPARAMETERS_LIST);

    /**
    *  Loads the code-generator with the proper CDML target file.
    *  It doesn't load it, if it has been loaded before.
    *  @param pkcTARGET Target code generator.
    */
    void loadCodeGenerator (const char* pkcTARGET);

    /**
    *  Writes on output, the code corresponding to instruction
    *  \a pkcINSTRUCTION_NAME and clause \a pkcCLAUSE_NAME.
    *  @param pkcINSTRUCTION_NAME Instruction name.
    *  @param pkcCLAUSE_NAME      Clause name.
    */
    void outputClauseCode (const char* pkcINSTRUCTION_NAME, const char* pkcCLAUSE_NAME = "body");

    /**
    *  Appends on output, \a zLENGTH characters from \a pkcSOURCE_TEXT.
    *  @param pkccSOURCE_TEXT Source text.
    *  @param zLENGTH         Source text length.
    */
    void outputCode (const char* pkccSOURCE_TEXT, size_t zLENGTH = 0);

    /**
    *  Appends on output, the text in \a rkySOURCE_TEXT.
    *  @param rkySOURCE_TEXT Source text.
    */
    void outputCode (const std::string& rkySOURCE_TEXT)
    {
      yOutputCode.append (rkySOURCE_TEXT);
    }

    /**
    *  Writes at the end of output, the text \a rkySOURCE_TEXT.
    *  @param rkySOURCE_TEXT Source text.
    */
    void postAppend (const std::string& rkySOURCE_TEXT);

    /**
    *  Writes at the begining of output, the text \a rkySOURCE_TEXT.
    *  @param rkySOURCE_TEXT Source text.
    */
    void preAppend (const std::string& rkySOURCE_TEXT);

    /**
    *  Sets the last read token from scanner.
    *  @param pkcLAST_TOKEN Last token read.
    */
    void pushTokenName (const char* pkcLAST_TOKEN);

    /**
    *  Sets the last read token from scanner.
    *  @param kyLAST_TOKEN Last token read.
    */
    void pushTokenName (const std::string& kyLAST_TOKEN);

    /**
    *  Splits the token \a pkcSOURCE_TEXT in, \a uiPERIOD_COUNT plus one,
    *  identifiers, and then, stored in the parser stack. It returns
    *  FOURTH_LEVEL_IDENTIFIER, THIRD_LEVEL_IDENTIFIER or SECOND_LEVEL_IDENTIFIER,
    *  corresponding to the number of periods inside the token.
    *  @param pkcSOURCE_TEXT Token name.
    *  @param uiPERIOD_COUNT Period count.
    *  @return Token code.
    */
    int splitToken (const char* pkcSOURCE_TEXT, unsigned int uiPERIOD_COUNT)
      throw (TExceptionUESQLC_Fatal);

    /**
    *  Opens files and runs the compiler.
    *  @return Exit status.
    */
    int start (void);


  public:

    //
    //  S E L E C T O R S
    //

    /**
    *  Create complementary files (if any). These files correspond with all
    *  variables named as "file::*".
    */
    void createComplementaryFiles (void) const;

    ///  Gets debug description, according with debug level.
    std::string debugDescription (void) const;

    /**
    *  Gets clause body for clause \a pkcCLAUSE_NAME in instruction
    *  \a pkcINSTRUCTION_NAME.
    *  @param pkcINSTRUCTION_NAME Instruction name.
    *  @param pkcCLAUSE_NAME      Clause name.
    *  @return Clause code.
    */
    std::string getClauseCode ( const char* pkcINSTRUCTION_NAME     ,
                                const char* pkcCLAUSE_NAME = "body" );

    /**
    *  Gets line directive for in line \a uiLINE_NUMBER. If
    *  \a uiLINE_NUMBER isn't passed, then it uses the line
    *  number of the last embedded introducer.
    *  @param uiLINE_NUMBER Line number.
    *  @return Text for line directive.
    */
    size_t getEmbeddedLineNumber (void) const
    {
      return zEmbeddedIntroducerLine;
    }

    /**
    *  Gets line directive for in line \a uiLINE_NUMBER. If
    *  \a uiLINE_NUMBER isn't passed, then it uses the line
    *  number of the last embedded introducer.
    *  @param uiLINE_NUMBER Line number.
    *  @return Text for line directive.
    */
    std::string getLineDirective (unsigned int uiLINE_NUMBER = 0) const;

    /**
    *  Handles syntactic errors.
    *  Throws an TExceptionUESQLC_Error exception with \a pkcMESSAGE message.
    *  @param pkcMESSAGE Error message.
    *  @exception TExceptionUESQLC_Error Exception for errors.
    */
    void handleCompilerError (const char* pkcMESSAGE) const;
    
    /**
    *  Returns true if current target has been used before.
    *  @return True if reusing target.
    */
    bool isReusingTarget (void) const;

    ///  Writes to standard error a debug description, according with debug level.
    void printDebugInformation (void) const
    {
      if ( uiCfgDebug > 0 )
      {
        cerr << debugDescription();
      }
    }

    /**
    *  Returns compiler program file name.
    *  @return Program file name.
    */
    const std::string programFileName (void) const;
    
    /**
    *  Verifies that invariant  corresponding  to \a liREQUERIMENT_IDENTIFIER,
    *  is true. If not, then throws the corresponding exception.
    *  @param liREQUERIMENT_IDENTIFIER Invariant identifier.
    *  \warning The proper declaration for this member function is:
    *           void require (long int, TStackValue*...);
    */
    void require (long int liREQUERIMENT_IDENTIFIER...) const;

    /**
    *  Returns true if compiler must use '#line' directives.
    *  @return True if using line directives.
    */
    bool useLineDirective (void) const;
    

  protected:

    //
    //  C O N S T R U C T O R S
    //

    ///  Opens files.
    void openFiles (void);

    ///  Closes files.
    void closeFiles (void);


  protected:

    //
    //  S E L E C T O R S
    //

    ///  Writes to standard error the exception description.
    void printExceptionDescription (const mpcl::TBaseException& rktSOURCE_EXCEPTION) const;

    ///  Writes to standard output the configuration parameters.
    void printParameters (void) const;

    ///  Writes to standard output the usage information.
    void printUsage (void) const;

};  // class TApplication


//
//  C O N S T R U C T O R S
//

inline void TApplication::
bindCodeParameter (const std::string& rkyVARIABLE, const std::string& rkyVALUE)
{

  tCodeParametersMap.bind (rkyVARIABLE, mpcl::TVariableString (rkyVALUE));

}  // bindCodeParameter()


//
//  E X P O R T E D   V A R I A B L E S
//

extern char    yytext[];
extern int     yylineno;
extern int     yydebug;
extern FILE*   yyin;
extern FILE*   yyout;


//
//  E X P O R T E D   F U N C T I O N S
//

extern int yyparse (void);


#endif  // not _UESQLC_APPLICATION__
